# -*- coding: utf-8 -*-
"""
This module contains classes and methods to handle lograptor configurations.
"""
#
# Copyright (C), 2011-2017, by SISSA - International School for Advanced Studies.
#
# This file is part of lograptor.
#
# Lograptor is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# file 'LICENSE' in the root directory of the present distribution
# for more details.
#
# @Author Davide Brunato <brunato@sissa.it>
#
import string
import socket
import re

try:
    import configparser
except ImportError:
    # Fall back for Python 2.x
    import ConfigParser as configparser
finally:
    RawConfigParser = configparser.RawConfigParser

from .exceptions import LogRaptorNoSectionError, LogRaptorNoOptionError, FileMissingError


class EnvInterpolation(object):
    _KEYCRE = re.compile(r"%\(([^)]+)\)s")

    def before_get(self, parser, section, option, value, defaults):
        L = []
        self._interpolate_some(parser, option, L, value, section, defaults, 1)
        return ''.join(L)

    def before_set(self, parser, section, option, value):
        tmp_value = value.replace('%%', '')  # escaped percent signs
        tmp_value = self._KEYCRE.sub('', tmp_value)  # valid syntax
        if '%' in tmp_value:
            raise ValueError("invalid interpolation syntax in %r at "
                             "position %d" % (value, tmp_value.find('%')))
        return value

    def before_read(self, parser, section, option, value):
        return value

    def before_write(self, parser, section, option, value):
        return value

    def _interpolate_some(self, parser, option, accum, rest, section, map, depth):
        rawval = parser.get(section, option, raw=True)
        if rawval:
            accum.append(string.Template(rawval).safe_substitute(parser.env))


class EnvConfigParser(RawConfigParser):
    """
    Environment-based configuration parser.

    It uses a 2-level dictionary for default values and to determine non-string
    option types. String type values are stripped by quotes and spaces. Configuration
    is determined by specific settings and complemented by a configuration file or a by
    default values. The env keyword arguments dictionary defines the values considered
    for the interpolation.

    :param settings: A two-level dictionary with customized sections and options.
    :param cfgfiles: A list of file paths where search for a configuration file or a \
    string with a path to a configuration file.
    :param defaults: A two-level dictionary with default settings.
    :param env: Environment values passed as keyword arguments.
    """
    _DEFAULT_INTERPOLATION = EnvInterpolation()
    optionxform = str  # Case sensitive option names

    _DEFAULTS = {}

    # noinspection PyMissingConstructor
    def __init__(self, settings=None, cfgfiles=None, defaults=None, **env):
        RawConfigParser.__init__(self, defaults=None, allow_no_value=False)
        self.__defaults = defaults if defaults is not None else self._DEFAULTS
        self.env = env
        self._interpolation = self._DEFAULT_INTERPOLATION   # Python 2.7 compatibility

        if cfgfiles:
            self.cfgfile = self.read_first(cfgfiles)
        else:
            self.cfgfile = None

        if settings is not None:
            self.read_dict(settings, 'settings')

    def __repr__(self):
        return u'%s(cfgfile=%r)' % (self.__class__.__name__, self.cfgfile)

    def defaults(self):
        return self.__defaults

    def get_default(self, section, option):
        try:
            return self.__defaults[section][option]
        except KeyError:
            return None

    def read_first(self, filenames):
        for filename in filenames if isinstance(filenames, (list, tuple)) else [filenames]:
            try:
                with open(filename) as fp:
                    try:
                        self.readfp(fp)
                    except AttributeError:
                        getattr(self, 'read_file')(fp)
            except IOError:
                pass
            else:
                return filename
        else:
            raise FileMissingError(
                "no configuration file in the list {} exists or is accessible!".format(filenames)
            )

    def get(self, section, option, default_section=None, raw=False, vars=None):
        if default_section is None:
            default_section = section

        used_defaults = False
        if section in self._sections:
            try:
                value = self._sections[section][option]
            except KeyError:
                try:
                    value = self.__defaults[default_section][option]
                    used_defaults = True
                except KeyError:
                    raise LogRaptorNoOptionError(option, section)

        elif default_section not in self.__defaults:
            raise LogRaptorNoSectionError(section)
        else:
            try:
                value = self.__defaults[default_section][option]
                used_defaults = True
            except KeyError:
                raise LogRaptorNoOptionError(option, section)

        if raw or value is None:
            return value
        elif used_defaults:
            options = self.options(section)
            return self._interpolation.before_get(self, default_section, option, value, options)
        else:
            options = self.options(section)
            return self._interpolation.before_get(self, section, option, value, options)

    def _get(self, section, conv, option, **kwargs):
        return conv(self.get(section, option))

    def getint(self, section, option, default_section=None):
        try:
            return RawConfigParser.getint(self, section, option)
        except configparser.NoOptionError:
            if default_section is None:
                raise
            return RawConfigParser.getint(self, default_section, option)

    def getfloat(self, section, option, default_section=None):
        try:
            return RawConfigParser.getfloat(self, section, option)
        except configparser.NoOptionError:
            if default_section is None:
                raise
            return RawConfigParser.getfloat(self, default_section, option)

    def getboolean(self, section, option, default_section=None):
        try:
            return RawConfigParser.getboolean(self, section, option)
        except configparser.NoOptionError:
            if default_section is None:
                raise
            return RawConfigParser.getboolean(self, default_section, option)

    def read_dict(self, dictionary, source='<dict>'):
        for section, options in dictionary.items():
            section = str(section)
            try:
                self.add_section(section)
            except (configparser.DuplicateSectionError, ValueError):
                pass

            for key, value in options.items():
                if value is not None:
                    default = self.get_default(section, key)
                    if isinstance(value, type(default)):
                        raise TypeError(' '.join([
                            "While reading from ", repr(source), ": wrong type for option ", repr(key),
                            " in section ", repr(section), ", ", repr(type(default)), "expected."
                        ]))
                    value = str(value)
                self.set(section, key, value)

    def options(self, section, prefix='', suffix='', default_section=None):
        try:
            opts = self._sections[section].copy()
        except KeyError:
            try:
                opts = self.__defaults[default_section or section].copy()
            except KeyError:
                raise LogRaptorNoSectionError(section)
        else:
            try:
                opts.update(self.__defaults[default_section or section])
            except KeyError:
                pass

        if '__name__' in opts:
            del opts['__name__']
        return list(filter(lambda x: x.startswith(prefix) and x.endswith(suffix), opts.keys()))

    def sections(self, prefix='', suffix=''):
        sections = self.__defaults.copy()
        sections.update(self._sections)
        return list(filter(lambda x: x.startswith(prefix) and x.endswith(suffix), sections.keys()))

    def items(self, section, raw=False, vars_=None):
        try:
            opts = self.__defaults[section].copy()
        except KeyError:
            opts = {}

        try:
            opts.update(self._sections[section])
        except KeyError:
            raise LogRaptorNoSectionError(section)

        if vars_:
            for key, value in vars_.items():
                opts[key] = value
        if '__name__' in opts:
            del opts['__name__']

        if raw:
            return [(k, opts[k]) for k in opts.keys()]
        else:
            return [
                (k, self._interpolation.before_get(self, section, k, opts[k], opts))
                for k in opts.keys()
            ]


class LogRaptorConfig(EnvConfigParser, object):
    _DEFAULTS = {
        'main': {
            'confdir': './conf.d/',
            'tmpdir': '/var/tmp/',
            'logdir': '/var/log/',
            'logfile': '/var/log/lograptor.log',
            'email_address': 'root@{0}'.format(socket.gethostname()),
            'smtp_server': '/usr/sbin/sendmail -t',
            'encodings': 'utf_8, latin1, latin2',
            'mapexp': 4,
        },
        'patterns': {
            'ASCII': r'[\x01-\x7f]*',
            'DNSNAME': r'\b(?:[A-Za-z0-9](?:[A-Za-z0-9-]*[A-Za-z0-9])?\.)*'
                       r'[A-Za-z0-9](?:[A-Za-z0-9-]*[A-Za-z0-9])?\b',
            'IPV4_ADDRESS': r'(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}'
                            r'(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)',
            'IPV6_ADDRESS': r'(?!.*::.*::)(?:(?!:)|:(?=:))(?:[0-9a-f]{0,4}(?:(?<=::)|(?<!::):)){6}'
                            r'(?:[0-9a-f]{0,4}(?:(?<=::)|(?<!::):)[0-9a-f]{0,4}'
                            r'(?: (?<=::)|(?<!:)|(?<=:) (?<!::) :)|'
                            r'(?:25[0-4]|2[0-4]\d|1\d\d|[1-9]?\d)'
                            r'(?: \.(?:25[0-4]|2[0-4]\d|1\d\d|[1-9]?\d)){3})',
            'USERNAME': r'[A-Za-z0-9!#$%&\'*+/=?^_`{|}~-]+(?:\.[A-Za-z0-9!#$%&\'*+/=?^_`{|}~-]+)*',
            'EMAIL': r'(?:|${USERNAME}|"${ASCII}")'
                     r'(?:|@(?:${DNSNAME}|\[(?:${IPV4_ADDRESS}|${IPV6_ADDRESS})\]))+',
            'ID': r'[0-9]+',
        },
        'fields': {
            'user': r'(|${USERNAME})',
            'mail': r'${EMAIL}',
            'from': r'${EMAIL}',
            'rcpt': r'${EMAIL}',
            'client': r'(${DNSNAME}|${IPV4_ADDRESS}|'
                      r'${DNSNAME}\[${IPV4_ADDRESS}\])',
            'pid': r'${ID}',
            'uid': r'${ID}',
            'msgid': r'${ASCII}',
        },
        # Reports
        'default_report': {
            'title': '${host} system events: ${localtime}',
            'html_template': './report_template.html',
            'text_template': './report_template.txt',
            'login_subreport': 'Logins',
            'mail_subreport': 'Mail report',
            'command_subreport': 'System commands',
            'query_subreport': 'Database lookups',
        },
        # Channels
        'stdout_channel': {
            'type': 'tty',
            'formats': 'text',
        },
        'mail_channel': {
            'type': 'mail',
            'formats': 'text, html, csv',
            'mailto': 'root',
            'include_rawlogs': 'False',
            'rawlogs_limit': '200',
            'gpg_encrypt': 'False',
            'gpg_keyringdir': '/root/.gnupg',
            'gpg_recipients': '',
            'gpg_signers': '',
        },
        'file_channel': {
            'type': 'file',
            'formats': 'text, html, csv',
            'notify': '',
            'pubdir': '/var/www/lograptor',
            'dirmask': '%Y-%b-%d_%a',
            'filemask': '%H%M',
            'save_rawlogs': 'False',
            'expire_in': '7',
            'pubroot': 'http://localhost/lograptor',
        }
    }

    def get_subreports(self, report):
        return [
            opt[:-10] for opt in self.options('%s_report' % report, suffix='_subreport')
        ]


class AppConfig(EnvConfigParser, object):
    _DEFAULTS = {
        'main': {
            'description': '${appname}',
            'tags': '${appname}',
            'sources': '${logdir}/messages',
            'enabled': True,
            'priority': 1,
        }
    }



# -*- coding: utf-8 -*-
"""
This module contains class to handle lograptor's lookup cache.
"""
#
# Copyright (C), 2011-2017, by SISSA - International School for Advanced Studies.
#
# This file is part of lograptor.
#
# Lograptor is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# file 'LICENSE' in the root directory of the present distribution
# for more details.
#
# @Author Davide Brunato <brunato@sissa.it>
#
import re
import socket
import string

try:
    import pwd
except ImportError:
    pwd = None


class LookupCache(object):
    """
    Name cache for names, that maps IPs to DNS names, UIDs to usernames.
    The names can be mapped into random generated values for obfuscate
    the input names, maintaining a correspondence for the entire process.
    """

    def __init__(self, args, config):
        self._maps = {}
        self.mapexp = config.getint('main', 'mapexp')
        self.mapmax = 10 ** self.mapexp
        self.ip_lookup = args.ip_lookup
        self.uid_lookup = args.uid_lookup
        self.anonymyze = args.anonymize
        self.fields = config.options('fields')
        self.base_gid_pattern = re.compile('^([a-zA-Z_]+)')
        ipv4_pattern = config.get('patterns', 'IPV4_ADDRESS')
        ipv6_pattern = config.get('patterns', 'IPV6_ADDRESS')
        self.ip_pattern = re.compile(u'({0}|{1})'.format(ipv4_pattern, ipv6_pattern))
        self.clear()

    def clear(self):
        self._maps.clear()
        for flt in set(self.fields) | {'host', 'thread', 'uid'}:
            self._maps[flt] = {}

    @property
    def hostsmap(self):
        return self._maps['host']

    @property
    def uidsmap(self):
        return self._maps['uid']

    def map_value(self, value, gid):
        """
        Return the value for a group id, applying requested mapping.
        Map only groups related to a filter, ie when the basename of
        the group is identical to the name of a filter.
        """
        base_gid = self.base_gid_pattern.search(gid).group(1)
        if self.anonymyze:
            try:
                if value in self._maps[base_gid]:
                    return self._maps[base_gid][value]
                else:
                    k = (len(self._maps[base_gid]) + 1) % self.mapmax
                    new_item = u'{0}_{1:0{2}d}'.format(base_gid.upper(), k, self.mapexp)
                    self._maps[base_gid][value] = new_item
                    return new_item
            except KeyError:
                return value
        elif base_gid in ['client', 'mail', 'from', 'rcpt', 'user'] and self.ip_lookup:
            ip_match = self.ip_pattern.search(value)
            if ip_match is None:
                return value
            host = self.gethost(ip_match.group(1))
            if host == ip_match.group(1) or value.startswith(host):
                return value
            return u''.join([
                value[:ip_match.start(1)],
                self.gethost(ip_match.group(1)),
                value[ip_match.end(1):]])
        elif (base_gid == 'user' or base_gid == 'uid') and self.uid_lookup:
            return self.getuname(value)
        else:
            return value

    def match_to_dict(self, match, gids):
        """
        Map values from match into a dictionary.
        """
        values = {}
        for gid in gids:
            try:
                values[gid] = self.map_value(match.group(gid), gid)
            except IndexError:
                pass
        return values

    def match_to_string(self, match, gids, values=None):
        """
        Return the mapped string from match object. If a dictionary of
        values is provided then use it to build the string.
        """
        s = match.string
        parts = []
        k = 0
        for gid in sorted(gids, key=lambda x: gids[x]):
            if values is None:
                try:
                    value = self.map_value(match.group(gid), gid)
                    parts.append(s[k:match.start(gid)])
                    parts.append(value)
                    k = match.end(gid)
                except IndexError:
                    continue
            elif gid in values:
                parts.append(s[k:match.start(gid)])
                parts.append(values[gid])
                k = match.end(gid)
        parts.append(s[k:])
        return u"".join(parts)

    def gethost(self, ip_addr):
        """
        Do reverse lookup on an ip address
        """
        # Handle silly fake ipv6 addresses
        try:
            if ip_addr[:7] == '::ffff:':
                ip_addr = ip_addr[7:]
        except TypeError:
            pass

        if ip_addr[0] in string.letters:
            return ip_addr

        try:
            return self.hostsmap[ip_addr]
        except KeyError:
            pass

        try:
            name = socket.gethostbyaddr(ip_addr)[0]
        except socket.error:
            name = ip_addr

        self.hostsmap[ip_addr] = name
        return name

    def getuname(self, uid):
        """
        Get the username of a given uid.
        """
        uid = int(uid)
        try:
            return self.uidsmap[uid]
        except KeyError:
            pass

        try:
            name = pwd.getpwuid(uid)[0]
        except (KeyError, AttributeError):
            name = "uid=%d" % uid

        self.uidsmap[uid] = name
        return name



# -*- coding: utf-8 -*-
"""
This module contains additional class and functions to handle time
and date values for lograptor package.
"""
#
# Copyright (C), 2011-2017, by SISSA - International School for Advanced Studies.
#
# This file is part of lograptor.
#
# Lograptor is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# file 'LICENSE' in the root directory of the present distribution
# for more details.
#
# @Author Davide Brunato <brunato@sissa.it>
#
import datetime
import re


DATE_FORMATS = (
    ('%%', re.compile(r"(%%)")),        # a literal %
    ('%a', re.compile(r"(?<!%)(%a)")),  # locale's abbreviated weekday name (e.g., Sun)
    ('%A', re.compile(r"(?<!%)(%A)")),  # locale's full weekday name (e.g., Sunday)
    ('%b', re.compile(r"(?<!%)(%b)")),  # locale's abbreviated month name (e.g., Jan)
    ('%B', re.compile(r"(?<!%)(%B)")),  # locale's full month name (e.g., January)
    ('%d', re.compile(r"(?<!%)(%d)")),  # day of month (e.g., 01)
    ('%j', re.compile(r"(?<!%)(%j)")),  # day of the year as zero padded decimal number (0 .. 366)
    ('%m', re.compile(r"(?<!%)(%m)")),  # month (01..12)
    ('%w', re.compile(r"(?<!%)(%w)")),  # day of the week (0 .. 6), 1 is monday
    ('%y', re.compile(r"(?<!%)(%y)")),  # last two digits of year (00..99)
    ('%Y', re.compile(r"(?<!%)(%Y)"))   # year
)


def parse_last_period(last):
    """
    Parse the --last value and return the time difference in seconds.
    """
    
    wordmap = {
        'hour': '1h',
        'day': '1d',
        'week': '1w',
        'month': '1m'
    }

    # seconds
    multmap = {
        'h': 3600,
        'd': 86400,
        'w': 604800,
        'm': 2592000
    }

    if last in wordmap:
        last = wordmap[last]

    cat = last[-1:].lower()

    if cat not in multmap:
        raise TypeError

    try: 
        num = int(last[:-1])
        if num <= 0:
            raise TypeError
    except ValueError: 
        raise TypeError
    
    diff = num * multmap[cat]    
    return diff


def get_datetime_interval(timestamp, diff, offset=0):
    """
    Returns datetime interval from timestamp backward in the past,
    computed using the milliseconds difference passed as argument.
    The final datetime is corrected with an optional offset.
    """
    fin_datetime = datetime.datetime.fromtimestamp(timestamp + offset)
    ini_datetime = datetime.datetime.fromtimestamp(timestamp - diff)
    return ini_datetime, fin_datetime


def parse_date_period(date):
    """
    Parse the --date value and return a couple of datetime object.
    The format is [YYYY]MMDD[,[YYYY]MMDD].
    """
    import datetime
    
    now = datetime.datetime.today()
    date_len = len(date)
    
    if date_len == 4:
        date1 = str(now.year) + date
        date2 = str(now.year) + date + "235959"
    elif date_len == 8:
        date1 = date
        date2 = date + "235959"
    elif date_len == 9:
        if date[4] != ',':
            raise TypeError
        date1 = str(now.year) + date[0:4]
        date2 = str(now.year) + date[5:9] + "235959"
    elif date_len == 17:
        if date[8] != ',':
            raise TypeError
        date1 = date[0:8]
        date2 = date[9:17] + "235959"
    else:
        raise TypeError

    try:
        date1 = datetime.datetime.strptime(date1, "%Y%m%d")
    except ValueError:
        if date_len < 9:
            raise ValueError("Error of date value in --date parameter, use --date=[YYYY]MMDD")
        else:
            raise ValueError("Error in the first date value in --date parameter, "
                             "use --date=[YYYY]MMDD,[YYYY]MMDD")
    try:
        date2 = datetime.datetime.strptime(date2, "%Y%m%d%H%M%S")        
    except ValueError:
        raise ValueError("Error in the second date value in --date parameter, "
                         "use --date=[YYYY]MMDD,[YYYY]MMDD")
        
    if date1 > date2:
        raise ValueError("Wrong parameter --date: the first date is after the second!")

    return date1, date2


class TimeRange(object):
    """
    A simple class to manage time range intervals
    """
    def __init__(self, time_range):
        """
        Constructor from timerange string.
        The time range format is HH:MM,HH:MM.
        """
        try:
            start_time, end_time = time_range.split(',')
        except ValueError:
            raise ValueError("%r is not a time range specification, use: HH:MM,HH:MM")

        self.start_time = datetime.datetime.strptime(start_time.strip(), '%H:%M').time()
        self.end_time = datetime.datetime.strptime(end_time.strip(), '%H:%M').time()
        if self.start_time == self.end_time:
            raise ValueError("times must be differents!")

        self.h1 = self.start_time.hour
        self.m1 = self.start_time.minute
        self.h2 = self.end_time.hour
        self.m2 = self.end_time.minute

    def between(self, time_):
        """
        Compare if the parameter HH:MM is in the time range.
        """
        hour = int(time_[0:2])
        minute = int(time_[3:5])
        return not (
            hour < self.h1 or hour > self.h2 or
            (hour == self.h1 and minute < self.m1) or
            (hour == self.h2 and minute > self.m2)
        )


def strftimegen(start_dt, end_dt):
    """
    Return a generator function for datetime format strings.
    The generator produce a day-by-day sequence starting from the first datetime
    to the second datetime argument.
    """
    if start_dt > end_dt:
        raise ValueError("the start datetime is after the end datetime: (%r,%r)" % (start_dt, end_dt))

    def iterftime(string):
        date_subs = [i for i in DATE_FORMATS if i[1].search(string) is not None]
        if not date_subs:
            yield string
        else:
            dt = start_dt
            date_path = string
            while end_dt >= dt:
                for item in date_subs:
                    date_path = item[1].sub(dt.strftime(item[0]), date_path)
                yield date_path
                dt = dt + datetime.timedelta(days=1)

    return iterftime



# -*- coding: utf-8 -*-
"""
This module define the matcher engine of lograptor package.
"""
#
# Copyright (C), 2011-2017, by SISSA - International School for Advanced Studies.
#
# This file is part of lograptor.
#
# Lograptor is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# file 'LICENSE' in the root directory of the present distribution
# for more details.
#
# @Author Davide Brunato <brunato@sissa.it>
#
import os
import time
import datetime
import logging
from collections import namedtuple, Counter


from .logparsers import CycleParsers
from .utils import open_resource

logger = logging.getLogger(__name__)


# Map for month field from any admitted representation to numeric.
MONTHMAP = {
    'Jan': '01', 'Feb': '02', 'Mar': '03', 'Apr': '04', 'May': '05', 'Jun': '06',
    'Jul': '07', 'Aug': '08', 'Sep': '09', 'Oct': '10', 'Nov': '11', 'Dec': '12',
    '01': '01', '02': '02', '03': '03', '04': '04', '05': '05', '06': '06',
    '07': '07', '08': '08', '09': '09', '10': '10', '11': '11', '12': '12'
}

NILVALUE = '-'  # RFC-5424 NILVALUE

MatcherResult = namedtuple(
    'MatcherResult', "lines matches unknown extra_tags first_event, last_event"
)


def get_mktime(year, month, day, ltime):
    try:
        return time.mktime((
            int(year),
            int(MONTHMAP[month]),
            int(day),
            int(ltime[:2]),     # Hour
            int(ltime[3:5]),    # Minute
            int(ltime[6:]),     # Second
            0, 0, -1
        ))
    except (KeyError, ValueError):
        return None


def get_mktime_period(time_period):
    try:
        start_dt = time.mktime(time_period[0].timetuple())
    except AttributeError:
        start_dt = float(0)
    try:
        return start_dt, time.mktime(time_period[1].timetuple())
    except AttributeError:
        return start_dt, time.mktime((2222, 2, 2, 0, 0, 0, 0, 0, 0))


def get_app(log_data, apps, tags, extra_tags):
    apptag = getattr(log_data, 'apptag', None)
    if apptag is None or apptag == NILVALUE or apptag.isdigit():
        # The app-tag is missing or has not a significative value
        for app in apps:
            if app.match_rules(log_data)[0]:
                return app
        return

    # Find app using the app-tag
    try:
        tag_apps = tags[apptag]
    except KeyError:
        tag_apps = [
            app for tag, _apps in tags.items() if apptag.startswith(tag)
            for app in _apps
        ]

    if not tag_apps:
        # Tag unmatched, skip the line
        extra_tags.update([apptag])
        return
    elif len(tag_apps) == 1:
        return tag_apps[0]
    else:
        # Match the tag using app's rules
        for app in tag_apps:
            if app.match_rules(log_data)[0]:
                return app
        else:
            # Return the first app (the line will be registered as unparsed)
            return tag_apps[0]


def create_search_function(invert, only_matching):
    def inverted_pattern_search(line, patterns):
        if not patterns:
            return False, None, line

        for regexp in patterns:
            match = regexp.search(line)
            if match is not None:
                return False, match, line
        else:
            return True, None, line

    def matching_pattern_search(line, patterns):
        if not patterns:
            return True, None, line

        for regexp in patterns:
            match = regexp.search(line)
            if match is not None:
                return True, match, '%s\n' % match.group(1)
        else:
            return False, None, line

    def normal_pattern_search(line, patterns):
        if not patterns:
            return True, None, line

        for regexp in patterns:
            match = regexp.search(line)
            if match is not None:
                return True, match, line
        else:
            return False, None, line

    if invert:
        return inverted_pattern_search
    elif only_matching:
        return matching_pattern_search
    else:
        return normal_pattern_search


host_cache = set()


def has_host_match(log_data, hosts):
    """
    Match the data with a list of hostname patterns. If the log line data
    doesn't include host information considers the line as matched.
    """
    hostname = getattr(log_data, 'host', None)
    if hostname and hostname not in host_cache:
        for host_pattern in hosts:
            if host_pattern.search(hostname) is not None:
                host_cache.add(hostname)
                return True
        else:
            return False
    return True


def create_matcher(dispatcher, parsers, apptags, matcher='ruled', hosts=tuple(), time_range=None,
                   time_period=(None, None), patterns=tuple(), invert=False, count=False,
                   files_with_match=None, max_count=0, only_matching=False, quiet=False,
                   thread=False, name_cache=None):
    """
    Create a matcher engine.
    :return: A matcher function.
    """
    parsers = CycleParsers(parsers)
    max_matches = 1 if quiet else max_count
    use_app_rules = matcher != 'unruled'
    select_unparsed = matcher == 'unparsed'
    register_log_lines = not (quiet or count or files_with_match is not None)
    start_dt, end_dt = get_mktime_period(time_period)
    pattern_search = create_search_function(invert, only_matching)
    dispatch_selected = dispatcher.dispatch_selected
    dispatch_context = dispatcher.dispatch_context

    def process_logfile(source, apps, encoding='utf-8'):
        log_parser = next(parsers)
        first_event = None
        last_event = None
        app_thread = None
        selected_data = None
        line_counter = 0
        unknown_counter = 0
        selected_counter = 0
        extra_tags = Counter()
        dispatcher.reset()

        with open_resource(source) as logfile:
            # Set counters and status
            logfile_name = logfile.name

            fstat = os.fstat(logfile.fileno())
            file_mtime = datetime.datetime.fromtimestamp(fstat.st_mtime)
            file_year = file_mtime.year
            file_month = file_mtime.month
            prev_year = file_year - 1

            for line in logfile:
                line = line.decode(encoding)
                line_counter += 1
                if line[-1] != '\n':
                    line += '\n'

                ###
                # Parses the line and extracts the log data
                log_match = log_parser.match(line)
                if log_match is None:
                    # The current parser doesn't match: try another available parser.
                    next_parser, log_match = parsers.detect(line)
                    if log_match is not None:
                        log_parser = next_parser
                    elif line_counter == 1:
                        logger.warning("the file '{}' has an unknown format, skip ...".format(logfile_name))
                        break
                    else:
                        unknown_counter += 1
                        continue
                log_data = log_parser.get_data(log_match)

                ###
                # Process last event repetition (eg. 'last message repeated N times' RFC 3164's logs)
                if getattr(log_data, 'repeat', None) is not None:
                    if selected_data is not None:
                        repeat = int(log_data.repeat)
                        if not thread:
                            selected_counter += repeat
                        if use_app_rules:
                            app = log_parser.app or get_app(selected_data, apps, apptags, extra_tags)
                            app.increase_last(repeat)
                            app.matches += 1
                            dispatch_context(
                                key=(app, app_thread),
                                filename=logfile_name,
                                line_number=line_counter,
                                rawlog=line
                            )
                        selected_data = None
                    continue
                selected_data = None

                ###
                # Parse the log's timestamp and gets the event datetime
                year = getattr(
                    log_data, 'year',
                    prev_year if MONTHMAP[log_data.month] != '01' and file_month == 1 else file_year
                )
                event_dt = get_mktime(
                    year=year,
                    month=log_data.month,
                    day=log_data.day,
                    ltime=log_data.ltime
                )

                ###
                # Scope exclusions
                if event_dt is None or event_dt < start_dt:
                    # Excludes lines older than the start datetime
                    continue
                elif event_dt > end_dt:
                    # Excludes lines newer than the end datetime
                    if fstat.st_mtime < event_dt:
                        logger.error("found anomaly with mtime of file %r at line %d", logfile_name, line_counter)
                    logger.warning("newer event at line %d: skip the rest of the file %r", line_counter, logfile_name)
                    break
                elif time_range is not None and not time_range.between(log_data.ltime):
                    # Excludes lines not in time range
                    continue
                elif hosts and not has_host_match(log_data, hosts):
                    # Excludes lines with host restriction
                    continue

                ###
                # Search log line with provided not-empty pattern(s)
                pattern_matched, match, rawlog = pattern_search(line, patterns)
                if not pattern_matched and not thread:
                    dispatch_context(filename=logfile_name, line_number=line_counter, rawlog=rawlog)
                    continue

                ###
                # App parsing: get the app from parser or from the log data
                app = log_parser.app or get_app(log_data, apps, apptags, extra_tags)
                if app is None:
                    # Unmatchable tag --> skip the line
                    continue
                elif use_app_rules:
                    # Parse the log message with app's rules
                    app_matched, has_full_match, app_thread, output_data = app.match_rules(log_data)
                    if not pattern_matched and app_matched and app_thread is None:
                        continue
                    if output_data:
                        rawlog = name_cache.match_to_string(log_match, log_parser.parser.groupindex, output_data)

                    if app_matched:
                        app.matches += 1
                        if not has_full_match or select_unparsed:
                            dispatch_context(
                                key=(app, app_thread),
                                filename=logfile_name,
                                line_number=line_counter,
                                rawlog=rawlog
                            )
                            continue
                    else:
                        app.unparsed += 1
                        if not select_unparsed:
                            dispatch_context(
                                key=(app, app_thread),
                                filename=logfile_name,
                                line_number=line_counter,
                                rawlog=rawlog
                            )
                            continue

                ###
                # Event selected: register event's data and datetime
                selected_data = log_data

                if first_event is None:
                    first_event = event_dt
                    last_event = event_dt
                else:
                    if first_event > event_dt:
                        first_event = event_dt
                    if last_event < event_dt:
                        last_event = event_dt

                if pattern_matched:
                    if max_matches and selected_counter >= max_matches:
                        # Stops iteration if max_count matches is exceeded
                        break
                    selected_counter += 1
                    if files_with_match:
                        break
                    if register_log_lines:
                        dispatch_selected(
                            key=(app, app_thread),
                            filename=logfile_name,
                            line_number=line_counter,
                            log_data=log_data,
                            rawlog=rawlog,
                            match=match
                        )
                elif register_log_lines and not only_matching:
                    # Thread matching
                    dispatch_context(
                        key=(app, app_thread),
                        filename=logfile_name,
                        line_number=line_counter,
                        rawlog=rawlog
                    )

        try:
            for key in list(dispatcher.keys()):
                dispatcher.flush(key)
        except (NameError, AttributeError):
            pass

        # If count option is enabled then register only the number of matched lines.
        if files_with_match and selected_counter or files_with_match is False and not selected_counter:
            dispatch_selected(filename=logfile.name)
        elif count:
            dispatch_selected(filename=logfile.name, counter=selected_counter)

        return MatcherResult(
            lines=line_counter,
            matches=selected_counter,
            unknown=unknown_counter,
            extra_tags=extra_tags,
            first_event=first_event,
            last_event=last_event
        )

    return process_logfile
